#include <stdexcept>
#include <msp/gl/extensions/arb_buffer_storage.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_map_buffer_range.h>
#include <msp/strings/format.h>
#include "buffer.h"
#include "error.h"
#include "misc.h"
#include "vertexsetup.h"

using namespace std;

namespace Msp {
namespace GL {

const Buffer *Buffer::bound[5] = { 0, 0, 0, 0, 0 };
BufferType buffer_types[] = { ARRAY_BUFFER, ELEMENT_ARRAY_BUFFER, PIXEL_PACK_BUFFER, PIXEL_UNPACK_BUFFER, UNIFORM_BUFFER };

Buffer::Buffer(BufferType t):
	type(t),
	size(0)
{
	require_buffer_type(type);

	if(ARB_direct_state_access)
		glCreateBuffers(1, &id);
	else
		glGenBuffers(1, &id);
}

Buffer::~Buffer()
{
	for(unsigned i=0; i<5; ++i)
		if(bound[i]==this)
			unbind_from(buffer_types[i]);
	glDeleteBuffers(1, &id);
}

void Buffer::require_buffer_type(BufferType type)
{
	static Require _req_vbo(ARB_vertex_buffer_object);
	if(type==PIXEL_PACK_BUFFER || type==PIXEL_UNPACK_BUFFER)
		static Require _req_pbo(ARB_pixel_buffer_object);
	else if(type==UNIFORM_BUFFER)
		static Require _req_ubo(ARB_uniform_buffer_object);
}

void Buffer::storage(unsigned sz)
{
	if(size>0)
		throw invalid_operation("Buffer::storage");
	if(sz==0)
		throw invalid_argument("Buffer::storage");

	size = sz;
	if(ARB_buffer_storage)
	{
		static const int flags = GL_MAP_READ_BIT|GL_MAP_WRITE_BIT|GL_DYNAMIC_STORAGE_BIT;
		if(ARB_direct_state_access)
			glNamedBufferStorage(id, size, 0, flags);
		else
		{
			BindRestore _bind(this, type);
			glBufferStorage(type, size, 0, flags);
		}
	}
}

void Buffer::set_usage(BufferUsage)
{
}

void Buffer::data(const void *d)
{
	if(size==0)
		throw invalid_operation("Buffer::data");

	if(ARB_buffer_storage)
		return sub_data(0, size, d);

	if(ARB_direct_state_access)
		glNamedBufferData(id, size, d, STATIC_DRAW);
	else
	{
		BindRestore _bind(this, type);
		glBufferData(type, size, d, STATIC_DRAW);
	}
}

void Buffer::data(unsigned sz, const void *d)
{
	if(size==0)
		storage(sz);
	else if(sz!=size)
		throw incompatible_data("Buffer::data");

	data(d);
}

void Buffer::sub_data(unsigned off, unsigned sz, const void *d)
{
	if(ARB_direct_state_access)
		glNamedBufferSubData(id, off, sz, d);
	else
	{
		BindRestore _bind(this, type);
		glBufferSubData(type, off, sz, d);
	}
}

void Buffer::require_size(unsigned req_sz) const
{
	if(size<req_sz)
		throw buffer_too_small(format("buffer has %d bytes; %d required", size, req_sz));
}

BufferRange *Buffer::create_range(unsigned s, unsigned o)
{
	return new BufferRange(*this, s, o);
}

void *Buffer::map()
{
	if(ARB_map_buffer_range)
	{
		if(ARB_direct_state_access)
			return glMapNamedBufferRange(id, 0, size, GL_MAP_READ_BIT|GL_MAP_WRITE_BIT);
		else
		{
			BindRestore _bind(this, type);
			return glMapBufferRange(type, 0, size, GL_MAP_READ_BIT|GL_MAP_WRITE_BIT);
		}
	}
	else if(ARB_direct_state_access)
		return glMapNamedBuffer(id, GL_READ_WRITE);
	else if(OES_mapbuffer)
	{
		BindRestore _bind(this, type);
		return glMapBuffer(type, GL_READ_WRITE);
	}
	else
		throw invalid_operation("Buffer::map");
}

bool Buffer::unmap()
{
	// TODO check if it's mapped
	if(ARB_direct_state_access)
		return glUnmapNamedBuffer(id);
	else if(OES_mapbuffer)
	{
		BindRestore _bind(this, type);
		return glUnmapBuffer(type);
	}
	else
		return true;
}

void Buffer::bind_to(BufferType t) const
{
	if(t!=type)
		require_buffer_type(t);
	if(t==ELEMENT_ARRAY_BUFFER)
		if(const VertexSetup *vs = VertexSetup::current())
		{
			// Don't change the binding in a vertex array object
			if(this==vs->get_index_buffer())
				return;
			throw invalid_operation("Buffer::bind_to(ELEMENT_ARRAY_BUFFER)");
		}
	if(set_current(t, this))
		glBindBuffer(t, id);
}

const Buffer *Buffer::current(BufferType t)
{
	if(t==ELEMENT_ARRAY_BUFFER)
		if(const VertexSetup *vs = VertexSetup::current())
			return vs->get_index_buffer();
	return binding(t);
}

void Buffer::unbind_from(BufferType type)
{
	if(type==ELEMENT_ARRAY_BUFFER && VertexSetup::current())
		throw invalid_operation("Buffer::unbind_from(ELEMENT_ARRAY_BUFFER)");
	if(set_current(type, 0))
		glBindBuffer(type, 0);
}

const Buffer *&Buffer::binding(BufferType type)
{
	switch(type)
	{
	case ARRAY_BUFFER:         return bound[0];
	case ELEMENT_ARRAY_BUFFER: return bound[1];
	case PIXEL_PACK_BUFFER:    return bound[2];
	case PIXEL_UNPACK_BUFFER:  return bound[3];
	case UNIFORM_BUFFER:       return bound[4];
	default: throw invalid_argument("Buffer::binding");
	}
}

bool Buffer::set_current(BufferType type, const Buffer *buf)
{
	const Buffer *&ptr = binding(type);
	if(ptr==buf)
		return false;

	ptr = buf;
	return true;
}


vector<const BufferRange *> BufferRange::bound_uniform;

BufferRange::BufferRange(Buffer &b, unsigned o, unsigned s):
	buffer(b),
	offset(o),
	size(s)
{
	if(o>buffer.get_size() || o+s>buffer.get_size())
		throw out_of_range("BufferRange::BufferRange");
}

BufferRange::~BufferRange()
{
	for(unsigned i=0; i<bound_uniform.size(); ++i)
		if(bound_uniform[i]==this)
			unbind_from(UNIFORM_BUFFER, i);
}

void BufferRange::data(const void *d)
{
	buffer.sub_data(offset, size, d);
}

void BufferRange::bind_to(BufferType t, unsigned i)
{
	if(t!=buffer.type)
		Buffer::require_buffer_type(t);
	if(set_current(t, i, this))
	{
		// The buffer gets bound as a side effect
		Buffer::set_current(t, &buffer);
		glBindBufferRange(t, i, buffer.id, offset, size);
	}
}

void BufferRange::unbind_from(BufferType t, unsigned i)
{
	if(set_current(t, i, 0))
	{
		Buffer::set_current(t, 0);
		glBindBufferBase(t, i, 0);
	}
}

const BufferRange *&BufferRange::binding(BufferType type, unsigned index)
{
	if(type==UNIFORM_BUFFER)
	{
		if(index>=get_n_uniform_buffer_bindings())
			throw out_of_range("BufferRange::binding");
		if(bound_uniform.size()<=index)
			bound_uniform.resize(index+1);
		return bound_uniform[index];
	}
	else
		throw invalid_argument("BufferRange::binding");
}

bool BufferRange::set_current(BufferType type, unsigned index, const BufferRange *buf)
{
	const BufferRange *&ptr = binding(type, index);
	if(ptr==buf)
		return false;

	ptr = buf;
	return true;
}

unsigned BufferRange::get_n_uniform_buffer_bindings()
{
	static unsigned count = get_i(GL_MAX_UNIFORM_BUFFER_BINDINGS);
	return count;
}

unsigned BufferRange::get_uniform_buffer_alignment()
{
	static unsigned align = get_i(GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT);
	return align;
}

} // namespace GL
} // namespace Msp
