#include <msp/core/hash.h>
#include <msp/strings/format.h>
#include "basicmaterial.h"
#include "gl.h"
#include "pbrmaterial.h"
#include "resources.h"
#include "texturing.h"
#include "uniform.h"
#include "unlitmaterial.h"

using namespace std;

namespace Msp {
namespace GL {

const Program *Material::create_compatible_shader(DataFile::Collection &coll) const
{
	string module_name;
	map<string, int> spec_values;
	fill_program_info(module_name, spec_values);

	string info = module_name;
	for(map<string, int>::const_iterator i=spec_values.begin(); i!=spec_values.end(); ++i)
		info += format(",%s:%d", i->first, i->second);

	string name = format("_material_%016x.shader", hash64(info));
	Program *shprog = coll.find<Program>(name);
	if(shprog)
		return shprog;

	const Module &module = coll.get<Module>(module_name);
	shprog = new Program(module, spec_values);
	try
	{
		coll.add(name, shprog);
	}
	catch(...)
	{
		delete shprog;
		throw;
	}

	return shprog;
}

void Material::attach_texture_to(const Texture *tex, Texturing &texturing, ProgramData &tex_shdata, const string &name) const
{
	if(!tex)
		return;

	int unit = -1;

	if(const Uniform *uni = tex_shdata.find_uniform(name))
		if(const Uniform1i *uni_int = dynamic_cast<const Uniform1i *>(uni))
			unit = uni_int->get();

	if(unit<0)
		unit = texturing.find_free_unit(name);
	if(unit<0)
		throw runtime_error("no free texunit");

	texturing.attach(unit, *tex, sampler);
	tex_shdata.uniform(name, unit);
}

Material::MaterialRegistry &Material::get_material_registry()
{
	static MaterialRegistry registry;
	static bool initialized = false;
	if(!initialized)
	{
		initialized = true;
		registry.register_type<BasicMaterial>("basic");
		registry.register_type<PbrMaterial>("pbr");
		registry.register_type<UnlitMaterial>("unlit");
	}
	return registry;
}


Material::Loader::Loader(Material &m):
	CollectionObjectLoader(m, 0)
{ }

Material::Loader::Loader(Material &m, Collection &c):
	CollectionObjectLoader(m, &c)
{ }

void Material::Loader::init_actions()
{
	add("sampler", &Loader::sampler);
}

void Material::Loader::sampler(const std::string &name)
{
	obj.sampler = &get_collection().get<Sampler>(name);
}


DataFile::Loader::ActionMap Material::GenericLoader::shared_actions;

Material::GenericLoader::GenericLoader(DataFile::Collection *c):
	coll(c)
{
	set_actions(shared_actions);
}

void Material::GenericLoader::init_actions()
{
	get_material_registry().add_all(*this);
}

} // namespace GL
} // namespace Msp
